<?php

namespace NoccyLabs\GdImage;

use GdImage;
use NoccyLabs\GdImage\Convolution\ConvolutionInterface;
use NoccyLabs\GdImage\Font\FontInterface;
use NoccyLabs\GdImage\Mixer\CopyMixer;
use NoccyLabs\GdImage\Mixer\MixerInterface;
use NoccyLabs\GdImage\Process\ProcessInterface;
use NoccyLabs\GdImage\Transformation\TransformationInterface;

class Image implements ImageInterface
{

    const ANCHOR_TOP_LEFT = 0;

    private int $width;

    private int $height;

    public GdImage $image;

    /**
     * Constructr
     * 
     * @param int|null $width
     * @param int|null $height
     * @param mixed $bgcolor
     * @throws \InvalidArgumentException 
     */
    public function __construct(int $width = null, int $height = null, int $bgcolor = 0xFFFFFF)
    {
        if ($width !== null && $height !== null) {
            $this->newImage($width, $height, $bgcolor);
        }
    }

    public function __clone()
    {
        $img = imagecreatetruecolor(imagesx($this->image), imagesy($this->image));
        imagecopy($img, $this->image, 0, 0, 0, 0, imagesx($this->image), imagesy($this->image));
        $this->image = $img;
    }

    public function newImage(int $width, int $height, int $bgcolor = 0xFFFFFF)
    {
        if ($width <= 0 || $height <= 0) {
            throw new \InvalidArgumentException("Invalid dimensions");
        }
        $this->image = imagecreatetruecolor($width, $height);
        imagefill($this->image, 0, 0, $bgcolor);
        $this->width = $width;
        $this->height = $height;
    }

    /**
     * Create a new image from an existing GdImage
     * 
     * @param GdImage $gd
     * @return Image
     */
    public static function fromGdImage(GdImage $gd): Image
    {
        $image = new Image();
        $image->image = $gd;
        $image->width = imagesx($gd);
        $image->height = imagesy($gd);
        return $image;
    }

    public static function fromFile(string $filename): Image
    {
        $image = new Image();
        $image->loadFromFile($filename);
        return $image;
    }

    public function loadFromFile(string $filename, array $opts = []): void
    {
        $extn = pathinfo($filename, PATHINFO_EXTENSION);
        switch (strtolower($extn)) {
            case 'png':
                $image = imagecreatefrompng($filename);
                break;
            case 'jpg':
            case 'jpeg':
                $image = imagecreatefromjpeg($filename);
                break;
            default:
                throw new \InvalidArgumentException("Unsupported file format for load");
        }

        if ($image??null) {
            $this->image = $image;
            $this->width = imagesx($image);
            $this->height = imagesy($image);
        }

    }

    public function saveToFile(string $filename, array $opts = []): void
    {
        $extn = pathinfo($filename, PATHINFO_EXTENSION);
        switch (strtolower($extn)) {
            case 'png':
                imagepng($this->image, $filename, intval($opts['pngquality'] ?? -1), intval($opts['pngfilters'] ?? -1));
                break;
            case 'jpg':
            case 'jpeg':
                imagejpeg($this->image, $filename, intval($opts['jpgquality'] ?? -1));
                break;
            default:
                throw new \InvalidArgumentException("Unsupported file format for save");
        }
    }

    public function getWidth(): int
    {
        return $this->width;
    }

    public function getHeight(): int
    {
        return $this->height;
    }

    /**
     * Draw an image into this image at the specified rectangle
     * 
     * @param Image $image
     * @param Rect $destRect
     * @param \NoccyLabs\GdImage\Mixer\MixerInterface|null $mixer
     * @return void
     */
    public function drawImage(Image $image, Rect $destRect, null|string|MixerInterface $mixer = null)
    {
        if (null === $mixer) {
            $mixer = new CopyMixer();
        } elseif (is_string($mixer)) {
            if (!class_exists($mixer)) {
                throw new \RuntimeException("Invalid mixer class");
            }
            $mixer = new $mixer;
        }

        // Create a temporary image and copy the source rect into it
        $destImage = $destRect->createImageFromRect();
        [$x, $y, $w, $h] = $destRect->getPoints();
        imagecopy($destImage->getGdImage(), $this->image, 0, 0, $x, $y, $w, $h);

        // Do the blending
        $mixer->blend($image->getGdImage(), $destImage->getGdImage(), $w, $h);

        // Copy the temporary image back into the source rect
        imagecopy($this->image, $destImage->getGdImage(), $x, $y, 0, 0, $w, $h);
    }

    public function drawLine(int $x1, int $y1, int $x2, int $y2, int $color)
    {
        imageline($this->image, $x1, $y1, $x2, $y2, $color);
    }

    public function drawPolyLine(array $points, int $color)
    {
        while (count($points) > 2) {
            $p1 = array_shift($points);
            $p2 = array_shift($points);
            $this->drawLine($p1[0], $p1[1], $p2[0], $p2[1], $color);
            array_unshift($points, $p2);
        }
    }

    public function drawRect(int $x1, int $y1, int $x2, int $y2, int $color)
    {
        imagerectangle($this->image, $x1, $y1, $x2, $y2, $color);
    }

    public function drawFilledRect(int $x1, int $y1, int $x2, int $y2, int $color)
    {
        imagefilledrectangle($this->image, $x1, $y1, $x2, $y2, $color);
    }


    /**
     * Draw text at the specific position in the image. Retrurns the resulting rect.
     * 
     * Some common options include:
     * 
     *  'outline' (null|int) outline color
     *  'shadow' (null|int) shdowccolor
     *  'anchor' (string) set align and position (default=topleft)
     * 
     * @param int $x
     * @param int $y
     * @param FontInterface $font
     * @param int $color
     * @param string $text
     * @param array $options
     * @return Rect
     */
    public function drawText(int $x, int $y, FontInterface $font, int $color, string $text, array $options=[]): Rect
    {
        $anchor = strtolower($options['anchor'] ?? 'topleft');

        [$tw, $th, $tx, $ty] = $font->measureString($text, $options);
        
        switch ($anchor) {
            case 'topleft':
                $ox = $x;
                $oy = $y;
                break;
            case 'topright':
                $ox = $x - $tw;
                $oy = $y;
                break;
            case 'bottomleft':
                $ox = $x;
                $oy = $y - $th;
                break;
            case 'bottomright':
                $ox = $x - $tw;
                $oy = $y - $th;
                break;
        }

        $font->renderString($this->image, $ox, $oy, $color, $text, $options);

        return new Rect($ox, $oy, $tw, $th);
    }

    /**
     * Apply a convolution filter to the image
     * 
     * @param ConvolutionInterface $convolution
     * @return void
     */
    public function convolution(ConvolutionInterface $convolution)
    {
        imageconvolution($this->image, $convolution->getMatrix(), $convolution->getDivisor(), $convolution->getOffset());
    }

    /**
     * Apply a convolution filter to the image directly from an array
     * 
     * @param array $matrix
     * @param float $divisor
     * @param float $offset
     * @return void
     */
    public function convolutionMatrix(array $matrix, float $divisor, float $offset)
    {
        imageconvolution($this->image, $matrix, $divisor, $offset);
    }

    public function transform(TransformationInterface $transformation, ?Rect $rect=null)
    {
        $transformation->applyTransform($this->image, $rect);
    }

    public function process(ProcessInterface $process, ?Rect $rect=null)
    {
        // TODO extract the rect if provided
        $process->applyProcess($this->image);
    }

    public function getRegion(Rect $rect): Image
    {
        $dest = $rect->createImageFromRect();

        [$l, $t, $w, $h] = $rect->getRect();
        imagecopy($dest->image, $this->image, 0, 0, $l, $t, $w, $h);

        return $dest;
    }

    public function getGdImage(): GdImage
    {
        return $this->image;
    }

}